from .pbstools import submit_jobs
import os
import string
import json
import shutil


def scatter(ndat, prefs, template_str):
    verbose = prefs["verbose"]

    # Get the template string
    template = string.Template(prefs[template_str]["template"])

    myjob = dict()

    save_files = list()

    if verbose:
        print("Printing jobs scatter jobs for " + template_str)

    # Slurm maxes out at ~32000 task array size
    # This changes how rows are selected by multiplying by tasks_per_job instead of creating a range using it
    reduced_range = int(ndat / prefs["job_prefs"]["tasks_per_job"]) + 1

    # For every cell
    for i in range(reduced_range):

        job_top = i * prefs["job_prefs"]["tasks_per_job"] + prefs["job_prefs"]["tasks_per_job"]
        if job_top > ndat:
            job_top = ndat

        job_range = range(i * prefs["job_prefs"]["tasks_per_job"], job_top)

        # Get the info required to build that job from the data spreadsheet
        myjob["prefs_path"] = prefs["my_path"]
        myjob["save_log_path"] = prefs["save_log_path"]
        myjob["row_index"] = " ".join([str(job_id) + " " for job_id in job_range])

        # Apply job info to the template
        outstr = prefs["prefix"] + "\n" + template.substitute(myjob)

        save_file = (
            prefs["job_prefs"]["script_dir"]
            + os.sep
            + template_str
            + "_"
            + str(i)
            + ".sh"
        )

        text_file = open(save_file, "w")
        text_file.write(outstr)
        text_file.close()

        save_files.append(save_file)

        tmp_file = "tmp_" + template_str + ".sh"

    return submit_jobs(save_files, prefs["job_prefs"], tmp_file_name=tmp_file)


def gather(prefs, template_str):
    verbose = prefs["verbose"]

    # Get the template string
    template = string.Template(prefs[template_str]["template"])

    if verbose:
        print("Printing " + template_str)

    # Get the info required to build that job from the data spreadsheet
    myjob = dict()
    myjob["prefs_path"] = prefs["my_path"]
    myjob["save_log_path"] = prefs["save_log_path"]

    # Apply job info to the template
    outstr = prefs["prefix"] + "\n" + template.substitute(myjob)

    save_file = prefs["job_prefs"]["script_dir"] + os.sep + template_str + ".sh"

    text_file = open(save_file, "w")
    text_file.write(outstr)
    text_file.close()

    tmp_file = "tmp_" + template_str + ".sh"

    return submit_jobs(save_file, prefs["job_prefs"], tmp_file_name=tmp_file)


def setup_prefs(json_path):
    with open(json_path) as f:
        prefs = json.load(f)

    #     pdb.set_trace()
    if "save_parent" not in prefs["global_vars"]:
        prefs["global_vars"]["save_parent"] = os.getcwd()

    prefs["save_parent"] = os.path.abspath(prefs["global_vars"]["save_parent"])

    with open(json_path, "w") as f:
        json.dump(prefs, f, indent=4, separators=(",", ": "))

    # make the parent directory if it doesnt exist
    if not os.path.exists(prefs["save_parent"]):
        os.makedirs(prefs["save_parent"])

    json_path_local = prefs["save_parent"] + os.sep + "prefs.json"
    if not os.path.exists(json_path_local):
        # make a copy of the json object in the parent directory
        shutil.copyfile(json_path, json_path_local)
    else:
        # use the local copy
        print("Local copy of preference file already exists at " + json_path_local)
        with open(json_path_local) as f:
            prefs = json.load(f)

    # record the location of the json object
    prefs["my_path"] = json_path_local
    # record the location of the data object
    prefs["save_log_path"] = prefs["save_parent"] + os.sep + prefs["data_log_name"]
    prefs["job_prefs"]["script_dir"] = (
        prefs["save_parent"] + os.sep + prefs["script_dir"]
    )

    if not os.path.exists(prefs["job_prefs"]["script_dir"]):
        os.makedirs(prefs["job_prefs"]["script_dir"])

    prefs["verbose"] = prefs["global_vars"]["verbose"]

    return prefs



import argparse
from scheduler_tools import Connector
import logging
import sys
import traceback

from pathlib import Path

###############################################################################

log = logging.getLogger()
logging.basicConfig(level=logging.INFO,
                    format='[%(levelname)4s:%(lineno)4s %(asctime)s] %(message)s')

###############################################################################

###############################################################################


class Args(argparse.Namespace):

    def __init__(self):
        # Arguments that could be passed in through the command line
        self.debug = False
        #
        self.__parse()

    # cluster = SLURMCluster(cores=2, memory="8GB", walltime="01:00:00", queue="aics_cpu_general")
    # cluster.adapt(minimum_jobs=2, maximum_jobs=40)

    def __parse(self):
        pth = Path("~/.aics_dask").expanduser()
        p = argparse.ArgumentParser(prog='spawn_adaptable_dask_cluster',
                                    description='launch a daemon to keep an adaptable dask cluster running for prefect')
        p.add_argument('-s', '--sshjson', help='json file with gateway, url, id, and identity file.', dest='sshjson',
                       type=Path, default=pth/"ssh.json")
        p.add_argument('-q', '--queue', dest='queue', type=str, default='aics_cpu_general',
                       help='The queue to use on the cluster {aics_cpu_general, aics_gpu_general, etc}.')

        p.add_argument('-c', '--cores', dest='cores', type=int, default=2,
                       help='number of cores to give each worker')
        p.add_argument('-g', '--memoryGB', dest='memory', type=int, default=8,
                       help='The amount of memory to request for each worker.')
        p.add_argument('-w', '--walltime', dest='walltime', type=str, default="02:00:00",
                       help='wall time for the worker nodes.')
        p.add_argument('-m', '--minimum_workers', dest='minworkers', type=int, default=2,
                       help='minimum number of workers to spin up.')
        p.add_argument('-x', '--maximum_workers', dest='maxworkers', type=int, default=40,
                       help='maximum number of workers to spin up.')
        p.add_argument('-r', '--remote_env', dest='remoteenv', type=str, default="dask-scheduler",
                       help='name of the remote environment to activate')
        p.add_argument('--debug', action='store_true', dest='debug', help=argparse.SUPPRESS)
        p.parse_args(namespace=self)


###############################################################################

def main():
    dbg = True
    try:
        args = Args()
        dbg = args.debug

        dask_prefs = dict()
        dask_prefs['queue'] = args.queue
        dask_prefs['cores'] = args.cores
        dask_prefs['ram'] = str(args.memory) + 'GB'
        dask_prefs['wall_time'] = args.walltime
        dask_prefs['min_workers'] = args.minworkers
        dask_prefs['max_workers'] = args.maxworkers
        dask_prefs['remote_env'] = args.remoteenv
        dask_prefs['remote_command'] = 'setup_and_spawn.bash'
        dask_prefs['remote_path'] = str(args.sshjson.parent.relative_to(Path().home()))
        print("remote path:", dask_prefs['remote_path'])
        ssh_info = args.sshjson

        conn = Connector(distributed_dict=dask_prefs, pref_path=ssh_info)
        conn.run_command()
        conn.stop_forward_if_running()
        conn.forward_ports()

    except Exception as e:
        log.error("=============================================")
        if dbg:
            log.error("\n\n" + traceback.format_exc())
            log.error("=============================================")
        log.error("\n\n" + str(e) + "\n")
        log.error("=============================================")
        sys.exit(1)


###############################################################################
# Allow caller to directly run this module (usually in development scenarios)


if __name__ == '__main__':
    main()




import asyncio
import asyncssh
import sys


class MySSHClientSession(asyncssh.SSHClientSession):
    def data_received(self, data, datatype):
        print(data, end='')

    def connection_lost(self, exc):
        if exc:
            print('SSH session error: ' + str(exc), file=sys.stderr)


class MySSHClient(asyncssh.SSHClient):
    def connection_made(self, conn):
        print('Connection made to %s.' % conn.get_extra_info('peername')[0])

    def auth_completed(self):
        print('Authentication successful.')


async def run_client():
    conn, client = await asyncssh.create_connection(MySSHClient, 'localhost')

    async with conn:
        chan, session = await conn.create_session(MySSHClientSession, 'ls abc')
        await chan.wait_closed()

try:
    asyncio.get_event_loop().run_until_complete(run_client())
except (OSError, asyncssh.Error) as exc:
    sys.exit('SSH connection failed: ' + str(exc))



import json
from scheduler_tools.types import Pathlike
from pathlib import Path
import getpass


class PrefectPreferences:
    """
    This class handles reading of a ~/.prefect/ssh.json file. This file has settings for the
    name of the gateway, the username to authenticate with, the path to the local ssh identity
    file.
    """

    def __init__(self, path: Pathlike = None):
        """

        :param path:
        """
        self._path = None

        if path is None:
            path = self.default_path() / "ssh.json"

        if not Path(path).expanduser().exists():
            raise FileNotFoundError

        self._path = Path(path).parent

        with open(path.expanduser(), "r") as f_in:
            self._data = json.load(f_in)

        self._check_json_obj()
        self._add_name_if_needed()
        self._add_known_hosts_if_needed()

    def default_path(self) -> Path:
        if self._path is None:
            print("loading default .aics_dask path")
            self._path = Path("~/.aics_dask")
        print(f"path: {self._path}")
        return self._path

    @property
    def gateway_url(self):
        return self._data['gateway']['url']

    @property
    def username(self):
        return self._data['gateway']['user']

    @property
    def identity_file(self):
        return self._data['gateway']['identityfile']

    @property
    def known_hosts(self):
        return self._data['known_hosts']

    def write_ssh_pid(self, pid):
        with open(str(self.ssh_pid_path()), 'w') as fp:
            fp.write(str(pid))

    def read_ssh_pid(self) -> [str, type(None)]:
        pid = None
        if self.ssh_pid_path().expanduser().exists():
            pid = open(str(self.ssh_pid_path().expanduser()), 'r').read()
        return pid

    def remove_ssh_pid(self):
        if self.ssh_pid_path().exists():
            self.ssh_pid_path().unlink()

    def ssh_pid_path(self):
        return self.default_path().expanduser() / "ssh_pid.txt"

    def cluster_job_id_path(self):
        return self.default_path().expanduser() / "cluster_job_id.txt"

    def read_prefect_job_id(self) -> [str, type(None)]:
        job_id = None
        if self.cluster_job_id_path().exists():
            job_id = open(str(self.cluster_job_id_path().expanduser()), 'r').read()
        return job_id

    def write_prefect_job_id(self, job_id):
        print(f"jobid: {job_id}")
        with open(str(self.cluster_job_id_path().expanduser()), 'w') as fp:
            fp.write(str(job_id))

    def remove_prefect_job_id(self):
        if self.cluster_job_id_path().exists():
            self.cluster_job_id_path().unlink()

    def cluster_pid_path(self):
        # this needs to be made dynamic
        return self.default_path().relative_to(Path().home()) / "pidfile"

    @property
    def local_dask_port(self):
        return self._data['dask_port']

    @property
    def local_dashboard_port(self):
        return self._data['dashboard_port']

    # private methods

    def _check_json_obj(self):
        assert "gateway" in self._data
        assert "url" in self._data["gateway"]
        assert "identityfile" in self._data["gateway"]
        # user can be optional and assume the local username extends to the gateway

    def _add_name_if_needed(self):
        if "user" not in self._data["gateway"]:
            self._data["gateway"]["user"] = getpass.getuser()

    def _add_known_hosts_if_needed(self):
        if "known_hosts" not in self._data:
            self._data["known_hosts"] = Path('~/.ssh/known_hosts').expanduser().resolve()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A simple example of a test file using a function.
NOTE: All test file names must have one of the two forms.
- `test_<XYY>.py`
- '<XYZ>_test.py'

Docs: https://docs.pytest.org/en/latest/
      https://docs.pytest.org/en/latest/goodpractices.html#conventions-for-python-test-discovery
"""

import pytest
from scheduler_tools import Example


# If you only have a single condition you need to test, a single test is _okay_ but parametrized tests are encouraged
def test_value_change():
    start_val = 5
    new_val = 20

    example = Example(start_val)
    example.update_value(new_val)
    assert (example.get_value() == new_val and example.get_previous_value() == start_val)


# Generally, you should parametrize your tests, but you should include exception tests like below!
@pytest.mark.parametrize("start_val, next_val, expected_values", [
    # (start_val, next_val)
    (5, 20, (20, 5)),
    (10, 40, (40, 10)),
    (1, 2, (2, 1))
])
def test_parameterized_value_change(start_val, next_val, expected_values):
    example = Example(start_val)
    example.update_value(next_val)
    assert expected_values == example.values


# The best practice would be to parametrize your tests, and include tests for any exceptions that would occur
@pytest.mark.parametrize("start_val, next_val, expected_values", [
    # (start_val, next_val)
    (5, 20, (20, 5)),
    (10, 40, (40, 10)),
    (1, 2, (2, 1)),
    pytest.param("hello", None, None, marks=pytest.mark.raises(exception=ValueError)),  # Init value isn't an integer
    pytest.param(1, "hello", None, marks=pytest.mark.raises(exception=ValueError))  # Update value isn't an integer
])
def test_parameterized_value_change_with_exceptions(start_val, next_val, expected_values):
    example = Example(start_val)
    example.update_value(next_val)
    assert expected_values == example.values



import pytest
from pathlib import Path
from scheduler_tools.connector import Connector
import asyncssh
import asyncio
from scheduler_tools.ScpDaskPrefs import run_launch_dask_head_node


class MockSaver(object):
    def write_prefect_job_id(self, job_id):
        print(f"mock saver asked to save cluster_job_id: {job_id}")


# test user name
def test_constructor():
    ctr = Connector()


async def remote_spawn():
    command = "sbatch -o ~/.aics_dask_gpu/sOut -e ~/.aics_dask_gpu/sErr -c 2 -p aics_gpu_general setup_and_spawn.bash dask_gpu ~/.aics_dask_gpu\n"
    command2 = "conda activate dask_gpu; echo 'Yes this seems to have executed'"
    async with asyncssh.connect('slurm-master.corp.alleninstitute.org',
                                client_keys=Path('/Users/jamies/Dropbox/SSH/AICS/privateKeys/jamiesAICS')) as conn:
        result = await conn.run(command, check=True)
        print(result)


def test_remote_spawn():
    asyncio.get_event_loop().run_until_complete(remote_spawn())


def test_remote_spawn_func():
    mock_saver = MockSaver()
    prefs = {
             'cores': 2,
             'queue': 'aics_gpu_general',
             'remote_env': 'dask_gpu',
             'remote_command': 'setup_and_spawn.bash'
             }
    print("sbatch -c 2 -p aics_gpu_general setup_and_spawn.bash dask_gpu ~/.aics_dask_gpu")
    asyncio.get_event_loop().run_until_complete(run_launch_dask_head_node('slurm-master.corp.alleninstitute.org',
                                                                          Path('/Users/jamies/Dropbox/SSH/AICS/privateKeys/jamiesAICS'),
                                                                          prefs,
                                                                          mock_saver,
                                                                          Path('.aics_dask_gpu')
                                                                          ))


def test_connect():
    ctr = Connector()
    ctr.create_connection(passphrase="")
    ctr.serialize_and_push_prefs()
    print("yes")







from dask_jobqueue import SLURMCluster
import dask, dask.distributed
from time import sleep
from datetime import datetime
from pathlib import Path

from aicsdaemon import Daemon

from .types import Pathlike


class AicsPrefectDaemon(Daemon):

    def __init__(self, slurm_prefs: dict, pidfile: Pathlike, stdin: Pathlike = None, stdout: Pathlike = None,
                 stderr: Pathlike = None, foreground: bool = False):
        super(AicsPrefectDaemon, self).__init__(pidfile=pidfile,
                                                stdout=stdout,
                                                stdin=stdin,
                                                stderr=stderr,
                                                foreground=foreground)
        print("AicsPrefectDaemon: pidfile: ", pidfile)
        open(stdout, 'w').close() if stdout else None
        self._slurm_prefs = slurm_prefs
        self._client = None
        self._localdir = Path(pidfile).parent / "logs" / datetime.now().strftime("%Y%m%d_%H:%M:%S")

    def run(self):

        cluster = SLURMCluster(cores=self._slurm_prefs['cores'],
                               memory=self._slurm_prefs['ram'],
                               walltime=self._slurm_prefs['wall_time'],
                               queue=self._slurm_prefs['queue'],
                               local_directory=self._localdir,
                               log_directory=self._localdir)
        cluster.adapt(minimum_jobs=self._slurm_prefs['min_workers'],
                      maximum_jobs=self._slurm_prefs['max_workers'])
        self._client = dask.distributed.Client(cluster)

        print(f"{cluster.scheduler_info['address']}")
        print(f"{cluster.scheduler_info['services']['dashboard']}", flush=True)

        # this enables us to put something in stdin and if so this loop will
        # exit. Thus using the stdin="filepath" argument gives us a clean way to
        # shutdown the process
        mtime = self.stdin.stat().st_mtime
        while mtime == self.stdin.stat().st_mtime:
            now = datetime.now()
            print(now)
            sleep(30)





import argparse
from scheduler_tools import Connector
import logging
import sys
import traceback

from pathlib import Path

###############################################################################

log = logging.getLogger()
logging.basicConfig(level=logging.INFO,
                    format='[%(levelname)4s:%(lineno)4s %(asctime)s] %(message)s')

###############################################################################

###############################################################################


class Args(argparse.Namespace):

    def __init__(self):
        # Arguments that could be passed in through the command line
        self.debug = False
        #
        self.__parse()

    # cluster = SLURMCluster(cores=2, memory="8GB", walltime="01:00:00", queue="aics_cpu_general")
    # cluster.adapt(minimum_jobs=2, maximum_jobs=40)

    def __parse(self):
        pth = Path("~/.aics_dask").expanduser()
        p = argparse.ArgumentParser(prog='shutdown_dask_cluster',
                                    description='stop remote dask head node and shut down port forwards.')
        p.add_argument('-s', '--sshjson', help='json file with gateway, url, id, and identity file.', dest='sshjson',
                       type=Path, default=pth / "ssh.json")
        p.add_argument('--debug', action='store_true', dest='debug', help=argparse.SUPPRESS)
        p.parse_args(namespace=self)


###############################################################################

def main():
    dbg = True
    try:
        args = Args()
        dbg = args.debug

        conn = Connector(pref_path=args.sshjson)
        conn.stop_dask()

    except Exception as e:
        log.error("=============================================")
        if dbg:
            log.error("\n\n" + traceback.format_exc())
            log.error("=============================================")
        log.error("\n\n" + str(e) + "\n")
        log.error("=============================================")
        sys.exit(1)


###############################################################################
# Allow caller to directly run this module (usually in development scenarios)


if __name__ == '__main__':
    main()

